/* $Id: call.c,v 1.7 2000/11/27 20:09:43 malekith Exp $ */

#include "compiler.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* func.c */
void check_base(CTX *ctx, struct class_def *c, const char *xclass, 
	        struct token *t);
/* expr.c */
void compile_exp(CTX *ctx, struct token *t, struct var_type *type);
int push_X(struct var_type *t);
void convert(CTX *ctx, struct token *t, struct var_type *from,
	     struct var_type *to);

struct class_func *find_func(CTX *ctx, struct token *t, struct token **args,
			     struct token **ptr)
{
	struct token *call, *name;
	const char *xclass, *fn;
	struct class_def *c;
	struct class_func *f;
	struct var_type x;
	
	x.ptr = 0;

	if (t->v_int == syn_pcall) {
		ctx->dont_emit++;
		compile_exp(ctx, t->left, &x);
		ctx->dont_emit--;
		if (x.id != sel_ptr) {
			oops(t,
			     "left side of `->' didn't result in a ptr");
			croak(ctx);
		}
		assert(x.ptr);
		*ptr = t->left;
		call = t->right;
	} else {
		call = t;
		*ptr = 0;
	}
		
	name = call->left;
	*args = call->right;
	xclass = 0;

	if (name->sel == sel_syn && name->v_int == syn_range) {
		assert(name->left->sel == sel_id);
		xclass = name->left->v_string;
		name = name->right;
	}
	assert(name->sel == sel_id);
	fn = name->v_string;
	c = x.ptr;

	if (xclass && strcmp(xclass, "Sys") == 0) {
		if (c) {
			oops(t, "cannot use ptr->Sys::XXX()");
			croak(ctx);
		} else
			goto sys;
	}

	/* if no class specified take `this' */
	if (c == 0)
		c = ctx->f->parent;

	if (xclass) {
		check_base(ctx, c, xclass, t);
		c = ph_find(class_hash, xclass);
	}

	for (; c; c = c->super)
		if ((f = ph_find(c->funcs, fn))) {
			f->ctx = x.ptr ? ctx_ptr : ctx_this;
			f->ctx_class = xclass ? c : 0;
			return f;
		}

	/* don't even try ptr->Sys::xxx() */
	if (x.ptr)
		goto no_sys;
		
      sys:
      	for (c = system_class; c; c = c->super)
      		if ((f = ph_find(c->funcs, fn))) {
			f->ctx = ctx_sys;
			f->ctx_class = system_class;
			return f;
		}
	
      no_sys:
	oops(t, "use of undefined function `%s' in %s::%s()",
	     fn, ctx->f->parent->name, ctx->f->name);
	croak(ctx);
}

static struct token *find_last_arg(struct token *t)
{
	struct token *r;
	
	if (t == 0)
		return 0;
		
	assert(t->sel == sel_syn);

	switch (t->v_int) {
	case syn_call_args:
		r = find_last_arg(t->right);
		if (r == 0)
			r = find_last_arg(t->left);
		break;
	case syn_call_arg:
		r = t;
		r->v_int = syn_call_arg_used;
		break;
	case syn_call_arg_used:
		r = 0;
		break;
	default:
		halt();
	}

	return r;
}

int push_X(struct var_type *t);

static void compute_args(CTX *ctx, struct var *arg, struct token *t, 
			 struct dbuf *stack)
{
	struct token *exp;
	struct var_type x;
	
	if (arg == 0)
		return;

	compute_args(ctx, arg->next, t, stack);
	exp = find_last_arg(t);
	if (exp == 0) {
		oops(t, "too few args for function call");
		croak(ctx);
	}
	
	exp = exp->left;

	compile_exp(ctx, exp, &x);

	/* now, we ought to convert x -> arg->type */
	convert(ctx, t, &x, &arg->type);

	emit(ctx,
		push_X(&arg->type),
		0
	);

	db_addc(stack, arg->type.id);
}

static void reset_args(struct token *t)
{
	if (t == 0)
		return;
	assert(t->sel == sel_syn);
	switch (t->v_int) {
	case syn_call_arg_used:
		t->v_int = syn_call_arg;
		break;
	case syn_call_arg:
		break;
	case syn_call_args:
		reset_args(t->left);
		reset_args(t->right);
		break;
	default:
		halt();
	}
}

static void do_call(CTX *ctx, struct class_func *f, struct token *ptr)
{
	struct var_type x;

	if (f->id == -1) {
		assert(ptr == 0);
		emit(ctx,
			op_mov_l, f->syscall,
			op_syscall,
			0
		);
		return;
	}
	
	switch (f->ctx) {
	case ctx_this:
		emit(ctx,
			op_mov_this_a,
			0
		);
		break;
	case ctx_sys:
		emit(ctx,
			op_mov_sys_a,
			0
		);
		break;
	case ctx_ptr:
		assert(ptr);
		compile_exp(ctx, ptr, &x);
		emit(ctx,
			op_mov_dA,
			0
		);
		break;
	default:
		halt();
	}

	if (f->ctx_class)
		emit(ctx,
			op_mov_l, f->id,
			op_jsr_id, f->ctx_class,
			0
		);
	else
		emit(ctx,
			op_mov_l, f->id,
			op_jsr,
			0
		);
}

static void clear_stack(CTX *ctx, struct dbuf *stack)
{
	int len, cnt;
	byte *f;

	f = db_finish(stack, &len);
	cnt = 0;

	if (len == 0)
		goto fini;
		
	emit(ctx,
		op_save_ds,
		0
	);
	
	while (len--)
		switch (f[len]) {
		case sel_ptr:
		case sel_int:
			cnt += sizeof(iptr);
			break;
		case sel_double:
			cnt += sizeof(real);
			break;
		case sel_string:
			if (cnt)
				emit(ctx,
					op_mov_l, cnt,
					op_sub_sp,
					0
				);
			cnt = 0;
			emit(ctx,
				op_pop_s,
				0
			);
		}
	if (cnt)
		emit(ctx,
			op_mov_l, cnt,
			op_sub_sp,
			0
		);

	emit(ctx,
		op_load_ds,
		0
	);
fini:
	free(f);
}

void compile_call(CTX *ctx, struct token *t, struct var_type *type)
{
	struct token *args, *ptr;
	struct class_func *f;
	struct dbuf *stack;

	f = find_func(ctx, t, &args, &ptr);
	reset_args(args);
	stack = db_new();
	compute_args(ctx, f->args, args, stack);
	if (find_last_arg(args)) {
		oops(t, "too many arguments in call");
		croak(ctx);
	}
	do_call(ctx, f, ptr);
	clear_stack(ctx, stack);
	copy_type(type, &f->type);
}
