/* $Id: expr.c,v 1.9 2000/11/27 20:09:44 malekith Exp $ */

#include "compiler.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* call.c */
void compile_call(CTX *ctx, struct token *t, struct var_type *type);

/* control.c */
void compile_stmt(CTX *ctx, struct token *t);

/* func.c */
void print_exp(struct token *t);

/* expr.c */
void compile_exp(CTX *ctx, struct token *t, struct var_type *type);
static void compile_lval(CTX *ctx, struct token *t, struct var_type *type);

/* newdel.c */
void compile_new(CTX *ctx, struct token *t, struct var_type *type);
void compile_delete(CTX *ctx, struct token *t, struct var_type *type);

extern int debug;
enum {
	f_allow_float	= 0x0001,
	f_allow_string	= 0x0002,
	f_allow_ptr	= 0x0004,
	f_one_arg	= 0x0008,
	f_need_lval	= 0x0010,
	f_need_bools	= 0x0020,
	f_result_bool	= 0x0040,
	f_assign	= 0x0080,
	f_pre		= 0x0100,
	f_lazy		= 0x0200,
	f_dont_swap	= 0x0400
};

#define	f_cmp	(f_allow_float | f_allow_string | f_result_bool)
#define NOP 31337
struct expr_op {
	int tok;
	int flags;
	int op_d, op_f, op_s;
};

struct expr_op ops[] = {
	{ 
		'+', 
		f_allow_float |
		f_allow_string,
		op_add_d,
		op_add_f,
		op_add_s
	}, {
		'-',
		f_allow_float,
		op_sub_d,
		op_sub_f,
		0
	}, {
		'*',
		f_allow_float,
		op_mul_d,
		op_mul_f,
		0
	}, {
		'/',
		f_allow_float,
		op_div_d,
		op_div_f,
		0
	}, {
		'%',
		0,
		op_mod_d,
		0,
		0
	}, {
		'|',
		0,
		op_or_d,
		0,
		0
	}, {
		'&',
		0,
		op_and_d,
		0,
		0
	}, {
		'^',
		0,
		op_eor_d,
		0,
		0
	}, {
		TOK_SL,
		0,
		op_shl_d,
		0,
		0
	}, {
		TOK_SR,
		0,
		op_shr_d,
		0,
		0
	}, {
		TOK_2_AND,
		f_allow_float 	|
		f_allow_ptr 	|
		f_need_bools 	|
		f_result_bool	|
		f_lazy,
		op_and,
		0,
		0
	}, {
		TOK_2_OR,
		f_allow_float	|
		f_allow_ptr 	|
		f_need_bools 	|
		f_result_bool	|
		f_lazy,
		op_or,
		0,
		0
	}, {
		TOK_2_XOR,
		f_allow_float 	|
		f_allow_ptr 	|
		f_need_bools 	|
		f_result_bool,
		op_eor,
		0,
		0
	}, {
		TOK_EQ,
		f_cmp | f_allow_ptr, 
		op_teq_d,
		op_teq_f,
		op_teq_s
	}, {
		TOK_NE,
		f_cmp | f_allow_ptr, 
		op_tne_d,
		op_tne_f,
		op_tne_s
	}, {
		TOK_LE,
		f_cmp, 
		op_tle_d,
		op_tle_f,
		op_tle_s
	}, {
		TOK_GE,
		f_cmp, 
		op_tge_d,
		op_tge_f,
		op_tge_s
	}, {
		'<',
		f_cmp, 
		op_tlt_d,
		op_tlt_f,
		op_tlt_s
	}, {
		'>',
		f_cmp, 
		op_tgt_d,
		op_tgt_f,
		op_tgt_s
	}, {
		TOK_PRE_2_PLUS,
		f_allow_float 	| 
		f_one_arg 	| 
		f_need_lval 	|
		f_pre, 
		op_inc_d,
		op_inc_f,
		0
	}, {
		TOK_PRE_2_MINUS,
		f_allow_float 	| 
		f_one_arg 	| 
		f_need_lval 	|
		f_pre, 
		op_dec_d,
		op_dec_f,
		0
	}, {
		TOK_POST_2_PLUS,
		f_allow_float 	| 
		f_one_arg 	| 
		f_need_lval, 
		op_inc_d,
		op_inc_f,
		0
	}, {
		TOK_POST_2_MINUS,
		f_allow_float 	| 
		f_one_arg 	| 
		f_need_lval, 
		op_dec_d,
		op_dec_f,
		0
	}, {
		'!',
		f_allow_float 	| 
		f_allow_ptr 	| 
		f_one_arg 	| 
		f_need_bools	|
		f_result_bool, 
		op_not,
		0,
		0
	}, {
		'~',
		f_one_arg, 
		op_not_d,
		0,
		0
	}, {
		TOK_NEG,
		f_allow_float |
		f_one_arg, 
		op_neg_d,
		op_neg_f,
		0
	}, {
		TOK_NOTNEG,
		f_allow_float |
		f_one_arg, 
		NOP,
		NOP,
		0
	}, {
		'=',
		f_allow_float	|
		f_allow_ptr	|
		f_allow_string	|
		f_assign, 
		NOP,
		NOP,	
		NOP
	}, {
		TOK_EPLUS,
		f_allow_float	|
		f_allow_string	|
		f_assign	|
		f_dont_swap, 
		op_add_d,
		op_add_f,
		op_add_s
	}, {
		TOK_EMINUS,
		f_allow_float	|
		f_assign, 
		op_sub_d,
		op_sub_f,
		0
	}, {
		TOK_ESLASH,
		f_allow_float	|
		f_assign, 
		op_div_d,
		op_div_f,
		0
	}, {
		TOK_ESTAR,
		f_dont_swap	|
		f_allow_float	|
		f_assign, 
		op_mul_d,
		op_mul_f,
		0
	}, {
		TOK_EAND,
		f_dont_swap	|
		f_assign, 
		op_and_d,
		0,
		0
	}, {
		TOK_EOR,
		f_dont_swap	|
		f_assign, 
		op_or_d,
		0,
		0
	}, {
		TOK_EXOR,
		f_dont_swap	|
		f_assign, 
		op_eor_d,
		0,
		0
	}, {
		TOK_ESR,
		f_assign, 
		op_shr_d,
		0,
		0
	}, {
		TOK_EPERC,
		f_assign, 
		op_div_d,
		0,
		0
	}, {
		TOK_ESL,
		f_assign, 
		op_shl_d,
		0,
		0
	}, {
		0,
		0,
		0,
		0,
		0
	}
};

void convert(CTX *ctx, struct token *t, struct var_type *from,
	     struct var_type *to)
{
	char *n;
	
	if (from->id == to->id)
		return;		/* nothing to do ... */
			
	if (from->id == sel_int && to->id == sel_double)
		emit(ctx, op_conv_df, 0);
	else if (from->id == sel_double && to->id == sel_int)
		emit(ctx, op_conv_fd, 0);
	else {
		n = strdup(name_of(from));
		oops(t, "don't know how to convert `%s' to `%s'",
			n, name_of(to));
		croak(ctx);
	}
}

static void check_type(CTX *ctx, struct expr_op *e, struct token *t, 
		       struct var_type *type)
{
	if (type->id == sel_double && !(e->flags & f_allow_float))
		oops(t, "float value not allowed here");
	else if (type->id == sel_ptr && !(e->flags & f_allow_ptr))
		oops(t, "ptr value not allowed here");
	else if (type->id == sel_string && !(e->flags & f_allow_string))
		oops(t, "string value not allowd here");
	else
		return;
	croak(ctx);
}

static int load_X(struct var_type *t)
{
	switch (t->id) {
	case sel_double:
		return op_mov_af;
	case sel_ptr:
	case sel_int:	
		return op_mov_ad;
	case sel_string:
		return op_mov_as;
	default:
		halt();
	}
}

static int swap_X(struct var_type *t)
{
	switch (t->id) {
	case sel_double:
		return op_swap_f;
	case sel_int:	
		return op_swap_d;
	case sel_string:
		return op_swap_s;
	default:
		halt();
	}
}

static int pop_X(struct var_type *t)
{
	switch (t->id) {
	case sel_double:
		return op_pop_f;
	case sel_ptr:
	case sel_int:	
		return op_pop_d;
	case sel_string:
		return op_pop_s;
	default:
		halt();
	}
}

int push_X(struct var_type *t)
{
	switch (t->id) {
	case sel_double:
		return op_push_f;
	case sel_ptr:
	case sel_int:	
		return op_push_d;
	case sel_string:
		return op_push_s;
	default:
		halt();
	}
}

static int stor_X(struct var_type *t)
{
	switch (t->id) {
	case sel_double:
		return op_mov_fa;
	case sel_ptr:
	case sel_int:	
		return op_mov_da;
	case sel_string:
		return op_mov_sa;
	default:
		halt();
	}
}

static int anti(int op)
{
	switch (op) {
	case op_dec_d:
		return op_inc_d;
	case op_dec_f:
		return op_inc_f;
	case op_inc_d:
		return op_dec_d;
	case op_inc_f:
		return op_dec_f;
	default:	
		halt();
	}
}

static int get_op(struct var_type *t, struct expr_op *e)
{
	int r;
	
	switch (t->id) {
	case sel_double:
		r = e->op_f;
		break;
	case sel_ptr:
	case sel_int:
		r = e->op_d;
		break;
	case sel_string:
		r = e->op_s;
		break;
	default:
		halt();
	}

	assert(r != 0);
	
	if (r == NOP)
		r = 0;

	return r;
}

static void convert_to_bool(CTX *ctx, struct var_type *t)
{
	switch (t->id) {
	case sel_int:
	case sel_ptr:
		break;
	case sel_double:
		emit(ctx,
			op_test_f,
			0
		);
		break;
	default:	
		halt();
	}
}

static void compile_assign(CTX * ctx, struct token *t, 
			   struct var_type *type, struct expr_op *e)
{
	struct var_type l, r;
	
	/* first of all we need to know types of left and right */
	ctx->dont_emit++;
	compile_lval(ctx, t->left, &l);
	compile_exp(ctx, t->right, &r);
	ctx->dont_emit--;
	/* next we shall check if they are ok: */
	check_type(ctx, e, t, &l);
	check_type(ctx, e, t, &r);
	if (l.id != r.id &&
	    !((l.id == sel_double || l.id == sel_int) &&
	      (r.id == sel_double || r.id == sel_int))) {
		oops(t, "cannot convert for assigment");
		croak(ctx);
	}
	/* FIXME: we should not allow assigning nor comparing
	 * ptrs of diffrent types ... */
	/* count the right side, and push converted result on the stack */
	compile_exp(ctx, t->right, &r);
	convert(ctx, t, &r, &l);
	emit(ctx,
		push_X(&l),
		0
	);

	/* now count the address of the right side */
	compile_lval(ctx, t->left, &l);
	
	if (debug && !ctx->dont_emit)
		print_exp(t);	/* print second time, just before the 
				 * very operation */

	if (get_op(&l, e)) {
		/* if we have sth to do (i.e. not a plain `=') */
		emit(ctx,
			load_X(&l),
			0
		);
		if (!(e->flags & f_dont_swap))
			emit(ctx,
				swap_X(&l),
				0
			);
		emit(ctx,
			get_op(&l, e),
			0
		);
	} else
		emit(ctx,
			pop_X(&l),
			0
		);
	emit(ctx,
		stor_X(&l),
		0
	);
	copy_type(type, &l);
}

static void compile_2arg(CTX * ctx, struct token *t, struct var_type *type,
			 struct expr_op *e)
{
	struct var_type l, r, target;
	int id = 0;

	/* first of all we need to know types of left and right */
	ctx->dont_emit++;
	compile_exp(ctx, t->left, &l);
	compile_exp(ctx, t->right, &r);
	ctx->dont_emit--;
	/* next we shall check if they are ok: */
	check_type(ctx, e, t, &l);
	check_type(ctx, e, t, &r);
	
	/* target of promotion */
	target.id = l.id;
	target.ptr = 0;

	if (e->flags & f_need_bools)
		target.id = sel_int;
		
	if (l.id == sel_double || r.id == sel_double)
		target.id = sel_double;
	
	if (!(e->flags & f_need_bools) && l.id != r.id &&
	    !((l.id == sel_double || l.id == sel_int) &&
	      (r.id == sel_double || r.id == sel_int))) {
		oops(t, "cannot convert");
		croak(ctx);
	}
	/* now, count left and optinally promote */
	compile_exp(ctx, t->left, &l);
	if (e->flags & f_need_bools) {
		convert_to_bool(ctx, &l);
		if (e->flags & f_lazy) {
			id = add_id(ctx);
			emit(ctx,
				(e->tok == TOK_2_OR) ? 
					op_btrue : op_bfalse, id,
				0
			);
		}
		
		emit(ctx, 
			op_push_d, 
			0
		);
	} else {
		convert(ctx, t, &l, &target);
		emit(ctx, push_X(&target), 0);
	}
	/* now, we have left on stack, try with right */
	compile_exp(ctx, t->right, &r);
	if (e->flags & f_need_bools)
		convert_to_bool(ctx, &r);
	else {
		convert(ctx, t, &r, &target);
		copy_type(&r, &target);
	}
	
	/* note that we allow PTR == PTR and PTR != PTR, it
	 * should be trated as int */
	if (r.id == sel_ptr)
		r.id = sel_int;
		
	if (debug && !ctx->dont_emit)
		print_exp(t);	/* print second time, just before the 
				 * very operation */
		
	emit(ctx, 
		get_op(&r, e), 
		0
	);
	if (e->flags & f_lazy)
		resolve(id);

	copy_type(type, (e->flags & f_result_bool) ? 0 : &r);
}

/* to be used from inside while and such */
void compute_logic(CTX *ctx, struct token *t)
{
	struct var_type x;

	compile_exp(ctx, t, &x);
	if (x.id == sel_string) {
		oops(t, "string value used in boolean context");
		croak(ctx);
	}
	convert_to_bool(ctx, &x);
}

/* 
 * load address of thing into %a 
 */
static void compile_lval(CTX *ctx, struct token *t, struct var_type *type)
{
	struct var *v;
	struct var_type x;
	/* 
	 * lval: rid | (ptr)->rid 
	 * rid: id | class :: id 
	 */

	if (t->sel == sel_id) {
		v = find_var(ctx, t->v_string, 0, 0, t);
	} else if (t->sel == sel_syn && t->v_int == syn_range) {
		v = find_var(ctx, t->right->v_string,
			     t->left->v_string, 0, t);
	} else if (t->sel == sel_syn && t->v_int == syn_pvar) {
		compile_exp(ctx, t->left, &x);
		if (x.id != sel_ptr) {
			oops(t, "left side of `->' didn't result in a ptr");
			croak(ctx);
		}
		assert(x.ptr);
		if (t->right->sel == sel_syn && t->right->v_int == syn_range)
			v = find_var(ctx, t->right->right->v_string, 
				     t->right->left->v_string, x.ptr, t);
		else if (t->right->sel == sel_id)
			v = find_var(ctx, t->right->v_string, 
				     0, x.ptr, t);
		else
			halt();
			
		emit(ctx, 
			op_mov_dA, 
			op_mov_l, v->offset,
			op_add_a,
			0
		);
		goto fini;
		
	} else {
		oops(t, "expression is not a proper lvalue");
		croak(ctx);
	}

	switch (v->ctx) {
	case ctx_arg:
		emit(ctx, 
			op_mov_l, -(v->offset + 3 * sizeof(iptr)),
			op_lda_bp,
			0
		);
		break;
	case ctx_local:	
		emit(ctx, 
			op_mov_l, v->offset,
			op_lda_bp,
			0
		);
		break;
	case ctx_this:
		emit(ctx, 
			op_mov_l, v->offset,
			op_lda_this,
			0
		);
		break;
	default:
		halt();
	}

fini:
	copy_type(type, &v->type);
}

static void compile_non_op(CTX *ctx, struct token *t, struct var_type *type)
{
	struct var_type x;
	
	if (t->sel == sel_id ||
	    (t->sel == sel_syn && (t->v_int == syn_pvar || 
	    			   t->v_int == syn_range))) {
		/* this is a lval, so also normal value */
		compile_lval(ctx, t, &x);
		copy_type(type, &x);
		emit(ctx,
			load_X(&x),
			0
		);
		return;
	}

	if (type) {
		type->ptr = 0;
		type->id = t->sel;
	}
	
	switch (t->sel) {
	case sel_int:
		emit(ctx,
			op_mov_l, t->v_int,
			0
		);
		return;
	case sel_double:
		emit(ctx,
			op_mov_f, t->v_double,
			0
		);
		return;
	case sel_string:
		emit(ctx,
			op_mov_s, t->v_string,
			0
		);
		return;
	case sel_ptr:
		assert(t->v_int == 0);
		emit(ctx,
			op_clr_d, 
			0
		);
		return;
	} 

	assert(t->sel == sel_syn);
	switch (t->v_int) {
	case syn_call:
	case syn_pcall:
		compile_call(ctx, t, type);
		break;
	case syn_delete:
		compile_delete(ctx, t->right, type);
		break;
	case syn_new:
		compile_new(ctx, t->right, type);
		break;
	default:
		oops(t, "not a proper expression");
		croak(ctx);
	}
}

/*
 * compile expression
 * if (type) { put result in %d|f|s and type in *type }
 */
void compile_exp(CTX *ctx, struct token *t, struct var_type *type)
{
	struct expr_op *e;
	struct var_type r;

	if (debug && !ctx->dont_emit)
		print_exp(t);
		
	if (t->sel != sel_op) {
		compile_non_op(ctx, t, type);
		return;
	}

	if (t->v_int == ',') {
		compile_exp(ctx, t->left, 0);
		compile_exp(ctx, t->right, type);
		return;
	}
	
	for (e = ops; e->tok; e++)
		if (t->v_int == e->tok)
			break;
	assert(e->tok != 0);

	if ((e->flags & f_need_lval) &&
	    (e->flags & f_one_arg)) {
		/* really special: ++/-- */
	    	compile_lval(ctx, t->right, &r);
	    	check_type(ctx, e, t, &r);
    		emit(ctx,
    			load_X(&r),
    			get_op(&r, e),
    			stor_X(&r),
    			/* if we were pre (`--a') or nobody is
    			 * intrested in result (`a--;'), don't
    			 * bother changing %d back to orignal
    			 * value */
    			(e->flags & f_pre) || type == 0 ? 
    				0 : 
    				anti(get_op(&r, e)),
    			0
    		);
    		copy_type(type, &r);
	} else if (e->flags & f_one_arg) {
	    	compile_exp(ctx, t->right, &r);
	    	check_type(ctx, e, t, &r);
	    	convert_to_bool(ctx, &r);
    		emit(ctx,
    			get_op(&r, e),
    			0
    		);
    		copy_type(type, 0);
	} else if (e->flags & f_assign) 
		compile_assign(ctx, t, type, e);
	else /* regular, 2-arg */
		compile_2arg(ctx, t, type, e);
}
