/* $Id: func.c,v 1.11 2000/11/27 20:09:44 malekith Exp $ */

#include "util.h"
#include "compiler.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern int debug;

struct func_def *func_head;

struct dbuf *code;
struct dbuf *class_xrefs;

static void dump(struct token *t)
{
	char *op = 0;

	if (t == 0)
		return;

	switch (t->sel) {
	case sel_int:
		printf("%d", t->v_int);
		break;
	case sel_double:
		printf("%f", t->v_double);
		break;
	case sel_id:
	case sel_string:
		printf("%s", t->v_string);
		break;
	case sel_syn:
		switch (t->v_int) {
		case syn_pvar:
			printf("(");
			dump(t->left);
			printf(")->(");
			dump(t->right);
			printf(")");
			break;
		case syn_range:
			dump(t->left);
			printf("::");
			dump(t->right);
			break;
		default:
			dump(t->left);
			printf(" ?SYN? ");
			dump(t->right);
			break;
		}
		break;
	case sel_op:
		switch (t->v_int) {
		case '-':
		case '+':
		case '*':
		case '/':
		case '&':
		case '|':
		case '^':
		case '%':
		case '=':
			printf("(");
			dump(t->left);
			printf(" %c ", t->v_int);
			dump(t->right);
			printf(")");
			break;
		case '~':
		case '!':
			printf("(%c", t->v_int);
			dump(t->right);
			printf(")");
			break;
			
		case TOK_POST_2_MINUS: 
			printf("("); 
			dump(t->right);
			printf("--)");
			break;
		case TOK_POST_2_PLUS:
			printf("("); 
			dump(t->right);
			printf("++)");
			break;
		case TOK_PRE_2_MINUS: 
			printf("(--"); 
			dump(t->right);
			printf(")");
			break;
		case TOK_PRE_2_PLUS:
			printf("(++"); 
			dump(t->right);
			printf(")");
			break;
			
		case TOK_NEG:
			printf("(-"); 
			dump(t->right);
			printf(")");
			break;
		case TOK_NOTNEG:
			printf("(+"); 
			dump(t->right);
			printf(")");
			break;
		
		case TOK_EPLUS: op = "+="; break;
		case TOK_EMINUS: op = "-="; break;
		case TOK_ESTAR: op = "*="; break;
		case TOK_ESLASH: op = "/="; break;
		case TOK_EPERC: op = "%="; break;
		case TOK_EXOR: op = "^="; break;
		case TOK_EOR: op = "|="; break;
		case TOK_EAND: op = "&="; break;
		case TOK_2_OR: op = "||"; break;
		case TOK_2_AND: op = "&&"; break;
		case TOK_2_XOR: op = "^^"; break;
		case TOK_SR: op = ">>"; break;
		case TOK_SL: op = "<<"; break;
		case TOK_EQ: op = "=="; break;
		case TOK_NE: op = "!="; break;
		case TOK_GE: op = ">="; break;
		case TOK_LE: op = "<="; break;
		}
		if (op) {
			printf("(");
			dump(t->left);
			printf(" %s ", op);
			dump(t->right);
			printf(")");
		}
		break;
	}
}

void print_exp(struct token *t)
{
	printf("# [");
	dump(t);
	printf("]\n");
}

void croak(CTX *ctx)
{
	longjmp(ctx->croak, 1);
}

int add_id(CTX *ctx)
{
	db_alloc(ctx->resolved, 4);
	return ctx->id++;
}

	
/* resolve_all() local */
static void finish_strings(CTX *ctx)
{
	u32 *p;
	byte *buf;
	u32 off;
	int n;

	off = db_off(ctx->code);
	 
	p = db_finish(ctx->strings, &n);
	memcpy(db_alloc(ctx->code, n), p, n);
	free(p);

	p = db_finish(ctx->strings_ids, &n);
	buf = (byte*)p;
	for (n /= 4; n--; p++)
		RES[*p] += off;
	free(buf);
}

static void resolve_all(CTX *ctx)
{
	int n0, n;
	u32 *res;
	u32 *offs, *p;
	u32 id, pc;
	byte *code;

	finish_strings(ctx);
	
	res = db_finish(ctx->resolved, 0);
	offs = db_finish(ctx->reloc, &n0);
	n0 /= 4;
	code = db_data(ctx->code);

	res[0] = ctx->sp;

	for (n = n0, p = offs; n--; p++) {
		id = *(u32*)(code + *p);
		pc = *p + 4;
		assert((id & 0x7fffffff) < (u32)ctx->id);
		if (id & 0x80000000)
			id = res[id & 0x7fffffff];
		else
			id = res[id] - pc;
		*(u32*)(code + *p) = id;
	}

	free(res);
	free(offs);
}

void check_base(CTX *ctx, struct class_def *c, const char *xclass, 
	        struct token *t)
{
	struct class_def *p;
	const char *n;

	if (xclass == 0)
		return;
		
	p = ph_find(class_hash, xclass);
	if (p == 0) {
		oops(t, "class `%s' undefined", xclass);
		croak(ctx);
	}
	n = c->name;
	for (; c; c = c->super)
		if (p == c)
			break;
	if (c == 0) {
		oops(t, "class `%s' is not a basetype for class `%s'",
		     xclass, n);
		croak(ctx);
	}
}

struct var *find_var(CTX *ctx, const char *name, const char *xclass,
		     struct class_def *c, struct token *t)
{
	struct var *v;
	
	if (xclass && strcmp(xclass, "Sys") == 0) {
		if (c) {
			oops(t, "cannot use ptr->Sys::XXX");
			croak(ctx);
		} else
			goto sys;
	}

	if (xclass || c)
		goto class_lookup;
		
	if ((v = ph_find(ctx->vars, name))) {
		v->ctx = ctx_local;
		return v;
	}
	
	for (v = ctx->f->args; v; v = v->next)
		if (strcmp(v->name, name) == 0)
			break;
	if (v) {
		v->ctx = ctx_arg;
		return v;
	}

class_lookup:
	/* if no class specified take `this' */
	if (c == 0)
		c = ctx->f->parent;
		
	if (xclass) {
		check_base(ctx, c, xclass, t);
		c = ph_find(class_hash, xclass);
	}
	
	for (; c; c = c->super)
		if ((v = ph_find(c->vars, name))) {
			v->ctx = ctx_this;
			return v;
		}

sys:
	for (c = system_class; c; c = c->super)
		if ((v = ph_find(c->vars, name))) {
			v->ctx = ctx_sys;
			return v;
		}

	oops(t, "use of undefined variable `%s' in %s::%s()",
	     name, ctx->f->parent->name, ctx->f->name);
	croak(ctx);
}

void compile_stmt(CTX *ctx, struct token *t);

/* do any special processing connected to constructor */
static void ctor(CTX *ctx)
{
	/* first of all, allocate space for us */
	emit(ctx,
		op_mov_l, ctx->f->parent->datasize,
		op_alloc, ctx->f->parent,
		0
	);
	
	/* if our class has any base class */
	if (ctx->f->parent->super)
		/* call its ctor */
		emit(ctx,
			op_mov_this_a,	/* %this -> %a */
			/* call func #0 (contructor) of base */
			op_clr_d,
			op_jsr_id, ctx->f->parent->super,
			0
		);
}

/* do any special processing connected to destructor */
static void dtor(CTX *ctx)
{
	/* if our class has any base class */
	if (ctx->f->parent->super)
		/* call its dtor */
		emit(ctx,
			op_mov_this_a,	/* %this -> %a */
			/* call func #1 (destructor) of base */
			op_inc_d,	/* shorter mov #1, %d */
			op_jsr_id, ctx->f->parent->super,
			0
		);
		
	/* commit suicide */
	emit(ctx,
		op_die,
		0
	);
}

/* certain symbols are emmited automagicly when some event occur,
 * ex: default dtor is emmited if none was defined and we
 * are compiling the ctor */
static void check_default_emit(CTX *ctx)
{
	struct token *t;
	
	if (ctx->f->parent->def_dtor && ctx->kind == kind_ctor) {
		/* generate empty:
		 * void ClassName::die() {}
		 * that looks kinda silly, i know ;) 
		 */
		t = new_token(
			new_token(new_token(0, 0, sel_kw, TOK_KW_VOID),
				  new_token(
				  	new_token(0, 0, sel_id, 
				  		ctx->f->parent->name),
				  	new_token(0, 0, sel_id, "die"),
				  	sel_syn, syn_func_name),
				  sel_syn, syn_func_info),
			new_token(0, 0, sel_syn, syn_func_body),
			sel_syn, 
			syn_func);
		if (compile_func(t))
			halt();	/* error?!?!? */
	}
}

unsigned int compile_block(struct token *t, int sp, struct func_def *f)
{
	struct compiler_state ctx;
	unsigned int r;
	
	if (code == 0)
		code = db_new();
	if (class_xrefs == 0)
		class_xrefs = db_new();

	r = db_off(code);
	ctx.code = code;
	ctx.vars = ph_new();
	(void)sp;
	ctx.sp = 0;
	ctx.dont_emit = 0;
	ctx.break_id = -1;
	ctx.id = 0;
	ctx.resolved = db_new();
	ctx.reloc = db_new();
	ctx.f = f;
	ctx.strings = db_new();
	ctx.strings_ids = db_new();
	ctx.class_xrefs = class_xrefs;

	if (strcmp(f->name, "die") == 0)
		ctx.kind = kind_dtor;
	else if (strcmp(f->name, f->parent->name) == 0)
		ctx.kind = kind_ctor;
	else
		ctx.kind = kind_norm;
	
	if (setjmp(ctx.croak))
		return -1;

	emit(&ctx, op_frame, add_id(&ctx) | 0x80000000, 0);
	if (ctx.kind == kind_ctor)
		ctor(&ctx);
	compile_stmt(&ctx, t);
	if (ctx.kind == kind_dtor)
		dtor(&ctx);
	else if (ctx.kind == kind_ctor)
		/* mov %this -> %a */
		emit(&ctx,
			op_mov_this_a,
			op_mov_Ad,
			0
		);
	emit(&ctx, op_rts, 0);
	resolve_all(&ctx);

	check_default_emit(&ctx);
	ph_std_kill(ctx.vars);
	return r;
}

int compile_func(struct token *t)
{
	struct func_def *f;
	struct class_func *proto;
	struct token *p;
	int sp = 0;
	
	f = NEW(struct func_def);
	memset(f, 0, sizeof(*f));

	p = t->left;
	SYN(p, func_info);
	
	if (token2type(&f->type, p->left))
		return -1;

	p = p->right;
	SYN(p, func_name);
	assert(p->left->sel == sel_id && p->right->sel == sel_id);
	f->name = take_string(p->right);
	
	f->parent = ph_find(class_hash, p->left->v_string);
	if (f->parent == 0) {
		oops(t, "%s::%s(): class undefined", p->left->v_string,
		     f->name);
		return 1;
	}

	proto = ph_find(f->parent->funcs, f->name);

	if (proto == 0) {
		oops(t, "no %s() function defined in class %s",
		     f->name, f->parent->name);
		return 1;
	}

	p = t->right;
	SYN(p, func_body);

	f->args = make_arg_list(p->left, &sp);
	f->id = proto->id;

	if (proto_args_diffrent(f->args, proto->args) ||
	    types_diffrent(&proto->type, &f->type)) {
		oops(t, "definition of %s::%s() doesn't match prototype",
		     f->parent->name, f->name);
		return 1;
	}
	
	if (debug) {
		struct var *a;
		
		printf("----> compiling function %s::%s\n", 
			f->parent->name, f->name);
		printf("%s %s::%s(", name_of(&f->type), 
			f->parent->name, f->name);
		for (a = f->args; a; a = a->next)
			printf("%s %s @%d%s", name_of(&a->type),
				a->name,
				a->offset,
				a->next ? ", " : "");
		printf(")\n");
	}

	f->block = compile_block(p->right, sp, f);
	if (f->block == (u32)-1)
		return -1;

	destroy_token(t);

	f->next = func_head;
	func_head = f;

	return 0;
}
